package dk.aau.cs.io;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import dk.aau.cs.model.tapn.*;
import dk.aau.cs.model.tapn.Colored.*;
import dk.aau.cs.model.tapn.Colored.Expressions.*;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import pipe.dataLayer.DataLayer;
import pipe.dataLayer.NetType;
import pipe.dataLayer.NetWriter;
import pipe.gui.MessengerImpl;
import pipe.gui.graphicElements.Arc;
import pipe.gui.graphicElements.Place;
import pipe.gui.graphicElements.Transition;
import pipe.gui.graphicElements.cpn.ColoredInputArcComponent;
import pipe.gui.graphicElements.cpn.ColoredOutputArcComponent;
import pipe.gui.graphicElements.tapn.TimedInhibitorArcComponent;
import pipe.gui.graphicElements.tapn.TimedOutputArcComponent;
import pipe.gui.graphicElements.tapn.TimedPlaceComponent;
import pipe.gui.graphicElements.tapn.TimedTransitionComponent;
import dk.aau.cs.util.Require;
import dk.aau.cs.verification.TAPNComposer;

public class PNMLWriter implements NetWriter {

	private TimedArcPetriNetNetwork network;
	private TimedArcPetriNet composedNetwork;
	private final HashMap<TimedArcPetriNet, DataLayer> guiModels;

	public PNMLWriter(
			TimedArcPetriNetNetwork network, 
			HashMap<TimedArcPetriNet, DataLayer> guiModels) {
		this.network = network;
		this.guiModels = guiModels;
	}
	
	public ByteArrayOutputStream savePNML() throws IOException, ParserConfigurationException, DOMException, TransformerConfigurationException, TransformerException {
		Document document = null;
		Transformer transformer = null;
		ByteArrayOutputStream os = new ByteArrayOutputStream();

		    TAPNComposer composer = new TAPNComposer(new MessengerImpl(), guiModels, true);
            composedNetwork = composer.transformModel(network).value1();

		// Build a Petri Net XML Document
		DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = builderFactory.newDocumentBuilder();
		document = builder.newDocument();

		Element pnmlRootNode = document.createElement("pnml"); // PNML Top Level
		document.appendChild(pnmlRootNode);
		pnmlRootNode.setAttribute("xmlns", "http://www.pnml.org/version-2009/grammar/pnml");

		Element netNode = document.createElement("net"); //Net node
		pnmlRootNode.appendChild(netNode);
		netNode.setAttribute("id", composedNetwork.name());
		netNode.setAttribute("type", "http://www.pnml.org/version-2009/grammar/ptnet");

        Element nameNode = document.createElement("name"); //Name of the net
        netNode.appendChild(nameNode);
        Element nameText = document.createElement("text");
        nameText.setTextContent(composedNetwork.name());
        nameNode.appendChild(nameText);
		
		Element pageNode = document.createElement("page"); //Page node
		netNode.appendChild(pageNode);
		pageNode.setAttribute("id", "page0");

		appendPlaces(document, composer.getGuiModel(), pageNode);
		appendTransitions(document, composer.getGuiModel(), pageNode);
		appendArcs(document, composer.getGuiModel(), pageNode);

		if(network.getNetType().equals(NetType.CPN)) {
			Element nameElement = document.createElement("name");
			netNode.appendChild(nameElement);
			Element nameTextElement = document.createElement("text");
			nameTextElement.setTextContent(composedNetwork.name());
			nameElement.appendChild(nameTextElement);
			appendDeclarations(document, composer.getGuiModel(), netNode);
		}

		document.normalize();
		// Create Transformer with XSL Source File
		transformer = TransformerFactory.newInstance().newTransformer();
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		DOMSource source = new DOMSource(document);

		StreamResult result = new StreamResult(os);
		transformer.transform(source, result);
		
		return os;
	}

	public void savePNML(File file) throws IOException, ParserConfigurationException, DOMException, TransformerConfigurationException, TransformerException {
		Require.that(file != null, "Error: file to save to was null");
		
		try {
			ByteArrayOutputStream os = savePNML();
			FileOutputStream fs = new FileOutputStream(file);
			fs.write(os.toByteArray());
			fs.close();
		} catch (ParserConfigurationException e) {
			System.out
					.println("ParserConfigurationException thrown in savePNML() "
							+ ": dataLayerWriter Class : dataLayer Package: filename=\"");
		} catch (DOMException e) {
			System.out
					.println("DOMException thrown in savePNML() "
							+ ": dataLayerWriter Class : dataLayer Package: filename=\""
							+ file.getCanonicalPath() + "\" transformer=\"");
		} catch (TransformerConfigurationException e) {
			System.out
					.println("TransformerConfigurationException thrown in savePNML() "
							+ ": dataLayerWriter Class : dataLayer Package: filename=\""
							+ file.getCanonicalPath()
							+ "\" transformer=\"");
		} catch (TransformerException e) {
			System.out
					.println("TransformerException thrown in savePNML() : dataLayerWriter Class : dataLayer Package: filename=\""
							+ file.getCanonicalPath()
							+ "\"" + e);
		}
	}

	private void appendPlaces(Document document, DataLayer guiModel, Element NET) {
		Place[] places = guiModel.getPlaces();
		for (int i = 0; i < places.length; i++) {
		    if (network.getNetType().equals(NetType.CPN)){
		        NET.appendChild(createColorPlaceElement((TimedPlaceComponent) places[i], guiModel, document));
            }
            else {
                NET.appendChild(createPlaceElement((TimedPlaceComponent) places[i], guiModel, document));
            }
		}
	}

	private void appendTransitions(Document document, DataLayer guiModel, Element NET) {
		Transition[] transitions = guiModel.getTransitions();
		for (int i = 0; i < transitions.length; i++) {
			if(network.getNetType().equals(NetType.CPN)) {
				NET.appendChild(createColoredTransitionElement((TimedTransitionComponent) transitions[i], document));
			}
			else {
				NET.appendChild(createTransitionElement((TimedTransitionComponent) transitions[i], document));
			}
		}
	}
	
	private void appendArcs(Document document, DataLayer guiModel, Element NET) {
		Arc[] arcs = guiModel.getArcs();
		for (int i = 0; i < arcs.length; i++) {
			if(network.getNetType().equals(NetType.CPN)) {
				NET.appendChild(createColoredArcElement(arcs[i],guiModel , document));
			}
			else {
				NET.appendChild(createArcElement(arcs[i], guiModel, document));
			}
		}
	}

	private Element createPlaceElement(TimedPlaceComponent inputPlace, DataLayer guiModel, Document document) {
		Require.that(inputPlace != null, "Error: inputPlace was null");
		Require.that(guiModel != null, "Error: guiModel was null");
		Require.that(document != null, "Error: document was null");

        Element placeElement = document.createElement("place");
        placeElement.setAttribute("id", (inputPlace.getId() != null ? inputPlace.getId() : inputPlace.getName()));

        Element name = document.createElement("name"); //Name
        placeElement.appendChild(name);
        Element nameGraphics = document.createElement("graphics");
        name.appendChild(nameGraphics);
        Element nameOffset = document.createElement("offset");
        nameGraphics.appendChild(nameOffset);
        nameOffset.setAttribute("x", (inputPlace.getNameOffsetXObject() != null ? String.valueOf(Math.round(inputPlace.getNameOffsetXObject())) : ""));
        nameOffset.setAttribute("y", (inputPlace.getNameOffsetYObject() != null ? String.valueOf(Math.round(inputPlace.getNameOffsetYObject())) : ""));
        Element nameText = document.createElement("text");
        name.appendChild(nameText);
        nameText.setTextContent(inputPlace.underlyingPlace().name());

        Element graphics = document.createElement("graphics");
        placeElement.appendChild(graphics);
        Element offset = document.createElement("position");
        graphics.appendChild(offset);
        offset.setAttribute("x", (inputPlace.getPositionXObject() != null ? String.valueOf(Math.round(inputPlace.getPositionXObject())) : ""));
        offset.setAttribute("y", (inputPlace.getPositionYObject() != null ? String.valueOf(Math.round(inputPlace.getPositionYObject()))	: ""));

        Element initialMarking = document.createElement("initialMarking"); //Name
        placeElement.appendChild(name);
        Element initialMarkingGraphics = document.createElement("graphics");
		initialMarking.appendChild(initialMarkingGraphics);
		Element initialMarkingOffset = document.createElement("offset");
		initialMarkingGraphics.appendChild(initialMarkingOffset);
		initialMarkingOffset.setAttribute("x", (inputPlace.getMarkingOffsetXObject() != null ? String.valueOf(Math.round(inputPlace.getMarkingOffsetXObject())) : ""));
		initialMarkingOffset.setAttribute("y", (inputPlace.getMarkingOffsetYObject() != null ? String.valueOf(Math.round(inputPlace.getMarkingOffsetYObject())) : ""));
		Element initialMarkingText = document.createElement("text");
		initialMarking.appendChild(initialMarkingText);
		initialMarkingText.setTextContent(((Integer) inputPlace.getNumberOfTokens() != null ? String.valueOf((Integer) inputPlace.getNumberOfTokens()) : "0"));
		placeElement.appendChild(initialMarking);

        return placeElement;
	}

	private Element createColorPlaceElement(TimedPlaceComponent inputPlace, DataLayer guiModel, Document document) {
        Require.that(inputPlace != null, "Error: inputPlace was null");
        Require.that(guiModel != null, "Error: guiModel was null");
        Require.that(document != null, "Error: document was null");

        if(inputPlace.underlyingPlace() instanceof LocalColoredPlace) {
				ColorType colorType = ((LocalColoredPlace) inputPlace.underlyingPlace()).getColorType();

			Element placeElement = document.createElement("place");
			placeElement.setAttribute("id", (inputPlace.getId() != null ? inputPlace.getId() : inputPlace.getName()));

			Element name = document.createElement("name"); //Name
			placeElement.appendChild(name);
			Element nameText = document.createElement("text");
			name.appendChild(nameText);
			nameText.setTextContent(inputPlace.underlyingPlace().name());
			Element nameGraphics = document.createElement("graphics");
			name.appendChild(nameGraphics);
			Element nameOffset = document.createElement("offset");
			nameGraphics.appendChild(nameOffset);
			nameOffset.setAttribute("x", (inputPlace.getNameOffsetXObject() != null ? String.valueOf(Math.round(inputPlace.getNameOffsetXObject())) : ""));
			nameOffset.setAttribute("y", (inputPlace.getNameOffsetYObject() != null ? String.valueOf(Math.round(inputPlace.getNameOffsetYObject())) : ""));

			Element graphics = document.createElement("graphics");
			placeElement.appendChild(graphics);
			Element offset = document.createElement("position");
			graphics.appendChild(offset);
			offset.setAttribute("x", (inputPlace.getPositionXObject() != null ? String.valueOf(Math.round(inputPlace.getPositionXObject())) : ""));
			offset.setAttribute("y", (inputPlace.getPositionYObject() != null ? String.valueOf(Math.round(inputPlace.getPositionYObject())) : ""));

			Element type = document.createElement("type");
			Element typeText = document.createElement("text");
			if(colorType != null) {
				if (colorType.getName().equals("dot")) {
					typeText.setTextContent("Dot");
				} else {
					typeText.setTextContent(colorType.getName());
				}
				type.appendChild(typeText);
				Element typeStructure = document.createElement("structure");
				type.appendChild(typeStructure);
				Element typeUsersort = document.createElement("usersort");
				typeUsersort.setAttribute("declaration", colorType.getId());

				typeStructure.appendChild(typeUsersort);
				placeElement.appendChild(type);
			}



			if(!inputPlace.underlyingPlace().tokens().isEmpty()) {
				Element hlInitialMarking = document.createElement("hlinitialMarking");
				Element hlInitialMarkingGraphics = document.createElement("graphics");
				hlInitialMarking.appendChild(hlInitialMarkingGraphics);
				Element hlInitialMarkingOffSet = document.createElement("offset");
				hlInitialMarkingGraphics.appendChild(hlInitialMarkingOffSet);
				hlInitialMarkingOffSet.setAttribute("x", (inputPlace.getMarkingOffsetXObject() != null ?
						String.valueOf(Math.round(inputPlace.getMarkingOffsetXObject())) : ""));
				hlInitialMarkingOffSet.setAttribute("y", (inputPlace.getMarkingOffsetYObject() != null ?
						String.valueOf(Math.round(inputPlace.getMarkingOffsetYObject())) : ""));
				Element hlInitialMarkingText = document.createElement("text");
				hlInitialMarking.appendChild(hlInitialMarkingText);
				String tokenNames = "";
				for (TimedToken token : inputPlace.underlyingPlace().tokens()) {
					ColoredToken ct = (ColoredToken) token;
					if (ct.getColor() instanceof DotConstant) {
						tokenNames += 1 + "'" + "(dot)";
					} else {
						String tokenName = ct.toStringForPNML();
						tokenNames += 1 + "'" + tokenName + " + ";
					}
				}

				hlInitialMarkingText.setTextContent(tokenNames);
				placeElement.appendChild(hlInitialMarking);
				Element hlStructure = document.createElement("structure");
				hlInitialMarking.appendChild(hlStructure);

				if (colorType.getName().equals("dot")) {
					Element dotNumberOf = document.createElement("numberof");
					hlStructure.appendChild(dotNumberOf);
					Element dotSubterm = document.createElement("subterm");
					dotNumberOf.appendChild(dotSubterm);
					Element dotNumbersConstsat = document.createElement("numberconstant");
					dotNumbersConstsat.setAttribute("value", "1");
					dotSubterm.appendChild(dotNumbersConstsat);
					Element dotPositive = document.createElement("positive");
					dotNumbersConstsat.appendChild(dotPositive);
					Element dotSubterm2 = document.createElement("subterm");
					dotNumberOf.appendChild(dotSubterm2);
					Element dotconstant = document.createElement("dotconstant");
					dotSubterm2.appendChild(dotconstant);
				}
				else {
					Element rangeAdd = document.createElement("add");
					hlStructure.appendChild(rangeAdd);
					for (TimedToken token : inputPlace.underlyingPlace().tokens()) {
						if(token instanceof ColoredToken) {
							Element rangeSubTerm = document.createElement("subterm");
							rangeAdd.appendChild(rangeSubTerm);
							Element rangeNumberOf = document.createElement("numberof");
							rangeSubTerm.appendChild(rangeNumberOf);
							Element rangeSubTerm2 = document.createElement("subterm");
							rangeNumberOf.appendChild(rangeSubTerm2);
							Element rangeNumberConstant = document.createElement("numberconstant");
							rangeNumberConstant.setAttribute("value", "1");
							rangeSubTerm2.appendChild(rangeNumberConstant);
							Element rangePositive = document.createElement("positive");
							rangeNumberConstant.appendChild(rangePositive);
							Element rangeSubTerm3 = document.createElement("subterm");
							rangeNumberOf.appendChild(rangeSubTerm3);

							if (colorType instanceof ProductType) {
								Element rangeTuble = document.createElement("tuple");
								rangeSubTerm3.appendChild(rangeTuble);
									for (Color element : ((ColoredToken) token).getColor().getTuple()) {
										Element rangeSubTerm4 = document.createElement("subterm");
										rangeTuble.appendChild(rangeSubTerm4);
										Element rangeUseroperator = document.createElement("useroperator");
										rangeUseroperator.setAttribute("declaration", element.getColorName());
										rangeSubTerm4.appendChild(rangeUseroperator);
									}
							} else {
								Element useroperatorElement = document.createElement("useroperator");
								useroperatorElement.setAttribute("declaration", ((ColoredToken)token).color().getName());
								rangeSubTerm3.appendChild(useroperatorElement);
							}
						}
					}
					return  placeElement;
				}
				return placeElement;
			}
			return placeElement;
		}
		return null;
    }

	private Element createTransitionElement(TimedTransitionComponent inputTransition, Document document) {
		Require.that(inputTransition != null, "Error: inputTransition was null");
		Require.that(document != null, "Error: document was null");
		
		Element transitionElement = document.createElement("transition");
		transitionElement.setAttribute("id", (inputTransition.getId() != null ? inputTransition.getId()	: "error"));

		Element name = document.createElement("name"); //Name
		transitionElement.appendChild(name);
		Element nameGraphics = document.createElement("graphics");
		name.appendChild(nameGraphics);
		Element nameOffset = document.createElement("offset");
		nameGraphics.appendChild(nameOffset);
		nameOffset.setAttribute("x", (inputTransition.getNameOffsetXObject() != null ? String.valueOf(Math.round(inputTransition.getNameOffsetXObject())) : ""));
		nameOffset.setAttribute("y", (inputTransition.getNameOffsetYObject() != null ? String.valueOf(Math.round(inputTransition.getNameOffsetYObject())) : ""));
		Element nameText = document.createElement("text");
		name.appendChild(nameText);
		nameText.setTextContent(inputTransition.underlyingTransition().name());
		
		Element graphics = document.createElement("graphics");
		transitionElement.appendChild(graphics);
		Element offset = document.createElement("position");
		graphics.appendChild(offset);
		offset.setAttribute("x", (inputTransition.getPositionXObject() != null ? String.valueOf(Math.round(inputTransition.getPositionXObject())) : ""));
		offset.setAttribute("y", (inputTransition.getPositionYObject() != null ? String.valueOf(Math.round(inputTransition.getPositionYObject())) : ""));

		return transitionElement;
	}

	private Element createColoredTransitionElement (TimedTransitionComponent inputTransition, Document document) {
		Require.that(inputTransition != null, "Error: inputTransition was null");
		Require.that(document != null, "Error: document was null");

		Element transitionElement = document.createElement("transition");
		transitionElement.setAttribute("id", (inputTransition.getId() != null ? inputTransition.getId()	: "error"));

		Element name = document.createElement("name"); //Name
		transitionElement.appendChild(name);
		Element nameGraphics = document.createElement("graphics");
		name.appendChild(nameGraphics);
		Element nameOffset = document.createElement("offset");
		nameGraphics.appendChild(nameOffset);
		nameOffset.setAttribute("x", (inputTransition.getNameOffsetXObject() != null ? String.valueOf(Math.round(inputTransition.getNameOffsetXObject())) : ""));
		nameOffset.setAttribute("y", (inputTransition.getNameOffsetYObject() != null ? String.valueOf(Math.round(inputTransition.getNameOffsetYObject())) : ""));
		Element nameText = document.createElement("text");
		name.appendChild(nameText);
		nameText.setTextContent(inputTransition.underlyingTransition().name());

		Element graphics = document.createElement("graphics");
		transitionElement.appendChild(graphics);
		Element offset = document.createElement("position");
		graphics.appendChild(offset);
		offset.setAttribute("x", (inputTransition.getPositionXObject() != null ? String.valueOf(Math.round(inputTransition.getPositionXObject())) : ""));
		offset.setAttribute("y", (inputTransition.getPositionYObject() != null ? String.valueOf(Math.round(inputTransition.getPositionYObject())) : ""));

		if(inputTransition.underlyingTransition() instanceof ColoredTransition) {
			Expression expr = ((ColoredTransition) inputTransition.underlyingTransition()).getGuard();
			if(expr != null) {
				Element conditionElement = document.createElement("condition");
				transitionElement.appendChild(conditionElement);
				Element textElement = document.createElement("text");
				String expressionName = ReplaceExpressionWithNames(expr.toString());
				textElement.setTextContent(expressionName);
				conditionElement.appendChild(textElement);
				Element structureElement = document.createElement("structure");
				conditionElement.appendChild(parseGuardExpression(expr, document, structureElement));
			}
		}
		return transitionElement;
	}

	private String ReplaceExpressionWithNames (String expression) {
		expression = expression.replace(">=", "gte");
		expression = expression.replace("<=", "lte");
		expression = expression.replace("<", "lt");
		expression = expression.replace(">", "gt");
		expression = expression.replace("=", "eq");
		expression = expression.replace("!=", "neq");
		return expression;
	}

	private Element parseGuardExpression(Expression expression, Document document, Element structureElement) {
		if(structureElement != null) {
			if (expression instanceof AndExpression) {
				Element andElement = document.createElement("and");
				Element subtermElement = document.createElement("subterm");
				andElement.appendChild(subtermElement);
				AndExpression expr = (AndExpression) expression;
				andElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
				Element subtermElement2 = document.createElement("subterm");
				andElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(andElement);
			} else if (expression instanceof OrExpression) {
				Element orElement = document.createElement("or");
				Element subtermElement = document.createElement("subterm");
				orElement.appendChild(subtermElement);
				OrExpression expr = (OrExpression) expression;
				orElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				orElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(orElement);
			} else if (expression instanceof EqualityExpression) {
				Element equalElement = document.createElement("equality");
				Element subtermElement = document.createElement("subterm");
				equalElement.appendChild(subtermElement);
				EqualityExpression expr = (EqualityExpression) expression;
				equalElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				equalElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(equalElement);
			} else if (expression instanceof GreaterThanEqExpression) {
				Element gteElement = document.createElement("greaterthanorequal");
				Element subtermElement = document.createElement("subterm");
				gteElement.appendChild(subtermElement);
				GreaterThanEqExpression expr = (GreaterThanEqExpression) expression;
				gteElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				gteElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(gteElement);
			} else if (expression instanceof GreaterThanExpression) {
				Element gtElement = document.createElement("greaterthan");
				Element subtermElement = document.createElement("subterm");
				gtElement.appendChild(subtermElement);
				GreaterThanExpression expr = (GreaterThanExpression) expression;
				gtElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				gtElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(gtElement);
			} else if (expression instanceof InequalityExpression) {
				Element ineuqlElement = document.createElement("inequality");
				Element subtermElement = document.createElement("subterm");
				ineuqlElement.appendChild(subtermElement);
				InequalityExpression expr = (InequalityExpression) expression;
				ineuqlElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				ineuqlElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(ineuqlElement);
			} else if (expression instanceof LessThanEqExpression) {
				Element lteElement = document.createElement("lessthanorequal");
				Element subtermElement = document.createElement("subterm");
				lteElement.appendChild(subtermElement);
				LessThanEqExpression expr = (LessThanEqExpression) expression;
				lteElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				lteElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(lteElement);
			} else if (expression instanceof LessThanExpression) {
				Element ltElement = document.createElement("lessthan");
				Element subtermElement = document.createElement("subterm");
				ltElement.appendChild(subtermElement);
				LessThanExpression expr = (LessThanExpression) expression;
				ltElement.appendChild(parseGuardExpression(expr.getLeftExpression(), document, subtermElement));
                Element subtermElement2 = document.createElement("subterm");
				ltElement.appendChild(parseGuardExpression(expr.getRightExpression(), document, subtermElement2));
				structureElement.appendChild(ltElement);
			} else if (expression instanceof NotExpression) {
				Element notElement = document.createElement("not");
				Element subtermElement = document.createElement("subterm");
				notElement.appendChild(subtermElement);
				NotExpression expr = (NotExpression) expression;
				notElement.appendChild(parseGuardExpression(expr.getExpression(), document, subtermElement));
				structureElement.appendChild(notElement);
			} else if (expression instanceof PredecessorExpression) {
				Element predElement = document.createElement("predecessor");
				Element subtermElement = document.createElement("subterm");
				predElement.appendChild(subtermElement);
				PredecessorExpression expr = (PredecessorExpression) expression;
				predElement.appendChild(parseGuardExpression(expr.getPredecessorExpression(), document, subtermElement));
				structureElement.appendChild(predElement);
			} else if (expression instanceof SuccessorExpression) {
				Element succElement = document.createElement("successor");
				Element subtermElement = document.createElement("subterm");
				succElement.appendChild(subtermElement);
				SuccessorExpression expr = (SuccessorExpression) expression;
				subtermElement.appendChild(parseGuardExpression(expr.getSuccessorExpression(), document, subtermElement));
				structureElement.appendChild(succElement);
			} else if (expression instanceof TupleExpression) {
				Element tupleElement = document.createElement("tuple");
				TupleExpression expr = (TupleExpression) expression;
				for (Expression colorExpressions : expr.getColors()) {
					Element subtermElement = document.createElement("subterm");
					tupleElement.appendChild(parseGuardExpression(colorExpressions, document, subtermElement));
				}
				structureElement.appendChild(tupleElement);
			} else if (expression instanceof UserOperatorExpression) {
				Element userOperationElement = document.createElement("useroperator");
				userOperationElement.setAttribute("declaration", ((UserOperatorExpression) expression).getUserOperator().getColorName());
				structureElement.appendChild(userOperationElement);
			} else if (expression instanceof VariableExpression) {
				Element variableElement = document.createElement("variable");
				variableElement.setAttribute("refvariable", ((VariableExpression) expression).getVariable().getId());
				structureElement.appendChild(variableElement);
			}
		}
		return structureElement;

	}

	private Element createArcElement(Arc arc, DataLayer guiModel, Document document) {
		Require.that(arc != null, "Error: inputArc was null");
		Require.that(guiModel != null, "Error: guiModel was null");
		Require.that(document != null, "Error: document was null");
		
		Element arcElement = document.createElement("arc");
		
		arcElement.setAttribute("id", (arc.getId() != null ? arc.getId() : "error"));
		arcElement.setAttribute("source", (arc.getSource().getId() != null ? arc.getSource().getId() : ""));
		arcElement.setAttribute("target", (arc.getTarget().getId() != null ? arc.getTarget().getId() : ""));
		
		if (arc instanceof TimedOutputArcComponent && ((TimedOutputArcComponent)arc).getWeight().value() > 1 ) {
			Element inscription = document.createElement("inscription");
			arcElement.appendChild(inscription);
			Element text = document.createElement("text");
			inscription.appendChild(text);
			text.setTextContent(((TimedOutputArcComponent)arc).getWeight().nameForSaving(false)+"");
		} 
		
		if(arc instanceof TimedInhibitorArcComponent){
			arcElement.setAttribute("type", "inhibitor");
		} else {
			arcElement.setAttribute("type", "normal");
		}
		
		
		//ArcPath
		int arcPoints = arc.getArcPath().getArcPathDetails().length;
		if (arcPoints > 2) {
			Element graphics = document.createElement("graphics");
			arcElement.appendChild(graphics);

			String[][] point = arc.getArcPath().getArcPathDetails();
			for (int j = 1; j < arcPoints-1; j++) { // Do not write the first and last point
				graphics.appendChild(createArcPoint(point[j][0],
						point[j][1], point[j][2], document, j));
			}
		}
		
		return arcElement;
	}

	private Element createArcExpressionElement(Arc arc, DataLayer guiModel, Document document, ArcExpression arcExpr) {

	    Element hlinscriptionElement = document.createElement("hlinscription");
		Element textElement = document.createElement("text");
		if(arc instanceof ColoredInputArcComponent)  {
			if(((ColoredInputArcComponent) arc).getArcExpression() != null) {
				textElement.setTextContent(((ColoredInputArcComponent) arc).getArcExpression().toString());
			}
		}
		else if(arc instanceof ColoredOutputArcComponent) {
			if(((ColoredOutputArcComponent) arc).getExpression() != null) {
				textElement.setTextContent(((ColoredOutputArcComponent) arc).getExpression().toString());
			}
		}
		hlinscriptionElement.appendChild(textElement);
		Element structureElement = document.createElement("structure");
		hlinscriptionElement.appendChild(structureElement);


		if( arc instanceof ColoredInputArcComponent) {
			ArcExpression expr = null;
			if( ((ColoredInputArcComponent) arc).getArcExpression() != null){
				ColoredInputArcComponent colorArc = (ColoredInputArcComponent)arc;
				expr = colorArc.getArcExpression();
				hlinscriptionElement.appendChild(parseArcExpression(expr, document, structureElement));
			}
		}
		else if (arc instanceof ColoredOutputArcComponent) {
			ArcExpression expr = null;
			if( ((ColoredOutputArcComponent) arc).getExpression() != null){
				ColoredOutputArcComponent colorArc = (ColoredOutputArcComponent)arc;
				expr = colorArc.getExpression();
				hlinscriptionElement.appendChild(parseArcExpression(expr, document, structureElement));
			}
		}
		return hlinscriptionElement;
	}

	private Element createColoredArcElement(Arc arc, DataLayer guiModel, Document document) {
		Require.that(arc != null, "Error: inputArc was null");
		Require.that(guiModel != null, "Error: guiModel was null");
		Require.that(document != null, "Error: document was null");

		Element arcElement = document.createElement("arc");
		arcElement.setAttribute("id", (arc.getId() != null ? arc.getId() : "error"));
		arcElement.setAttribute("source", (arc.getSource().getId() != null ? arc.getSource().getId() : ""));
		arcElement.setAttribute("target", (arc.getTarget().getId() != null ? arc.getTarget().getId() : ""));

		if(guiModel.netType().equals(NetType.CPN)) {
			ArcExpression arcExpr = null;
			if( arc instanceof ColoredInputArcComponent) {
				arcExpr = ((ColoredInputArc)((ColoredInputArcComponent) arc).underlyingTimedInputArc()).getArcExpression();
				if (arcExpr != null) {
					arcElement.appendChild(createArcExpressionElement(arc, guiModel, document, arcExpr));
				}
			} else if (arc instanceof  ColoredOutputArcComponent) {
				arcExpr = ((ColoredOutputArc)((ColoredOutputArcComponent) arc).underlyingArc()).getExpression();
				if (arcExpr != null) {
					arcElement.appendChild(createArcExpressionElement(arc, guiModel, document, arcExpr));
				}
			}
		}

		if (arc instanceof TimedOutputArcComponent && ((TimedOutputArcComponent)arc).getWeight().value() > 1 ) {
			Element inscription = document.createElement("inscription");
			arcElement.appendChild(inscription);
			Element text = document.createElement("text");
			inscription.appendChild(text);
			text.setTextContent(((TimedOutputArcComponent)arc).getWeight().nameForSaving(false)+"");
		}

		if(arc instanceof TimedInhibitorArcComponent){
			arcElement.setAttribute("type", "inhibitor");
		} else {
			arcElement.setAttribute("type", "normal");
		}

		//ArcPath
		int arcPoints = arc.getArcPath().getArcPathDetails().length;
		if (arcPoints > 2) {
			Element graphics = document.createElement("graphics");
			arcElement.appendChild(graphics);

			String[][] point = arc.getArcPath().getArcPathDetails();
			for (int j = 1; j < arcPoints-1; j++) { // Do not write the first and last point
				graphics.appendChild(createArcPoint(point[j][0],
						point[j][1], point[j][2], document, j));
			}
		}
		return arcElement;
	}

	private Element parseArcExpression(Expression expression, Document document, Element structureElement) {
		if(expression instanceof DotConstantExpression) {
			Element dotConstantElement = document.createElement("dotconstant");
			structureElement.appendChild(dotConstantElement);

		}
		else if (expression instanceof AndExpression) {
			Element andElement = document.createElement("and");
			Element subtermElement = document.createElement("subterm");
			andElement.appendChild(subtermElement);
			AndExpression expr = (AndExpression) expression;
			andElement.appendChild(parseArcExpression(expr.getRightExpression(), document, subtermElement));
			andElement.appendChild(parseArcExpression(expr.getLeftExpression(), document, subtermElement));
			structureElement.appendChild(andElement);

		}
		else if(expression instanceof NumberOfExpression) {
			Element numberOfElement = document.createElement("numberof");
			Element subtermElement = document.createElement("subterm");
			numberOfElement.appendChild(subtermElement);
			Element numberConstantElement = document.createElement("numberconstant");
			numberConstantElement.setAttribute("value", "1");
			subtermElement.appendChild(numberConstantElement);
			Element positiveElement = document.createElement("positive");
			numberConstantElement.appendChild(positiveElement);
			Element subtermElement2 = document.createElement("subterm");

			if(((NumberOfExpression) expression).getAllExpression() == null) {
				NumberOfExpression expr = (NumberOfExpression) expression;
				for (Expression colorExpression : expr.getNumberOfExpression()) {
					numberOfElement.appendChild(parseArcExpression(colorExpression, document, subtermElement2));
				}
			}
			else {
				Element allElement = document.createElement("all");
				subtermElement2.appendChild(allElement);
				Element usersortElement = document.createElement("usersort");
				allElement.appendChild(usersortElement);
				usersortElement.setAttribute("declaration", ((NumberOfExpression) expression).getAllExpression().getColorType().getName());
				numberOfElement.appendChild(allElement);
			}
			structureElement.appendChild(numberOfElement);
		}

		else if(expression instanceof UserOperatorExpression) {
			Element userOperationElement = document.createElement("useroperator");
			userOperationElement.setAttribute("declaration", ((UserOperatorExpression) expression).getUserOperator().getColorName());
			structureElement.appendChild(userOperationElement);
		}
		else if(expression instanceof VariableExpression) {
			Element variableElement = document.createElement("variable");
			variableElement.setAttribute("refvariable", ((VariableExpression) expression).getVariable().getId());
			structureElement.appendChild(variableElement);
		}
		else if (expression instanceof SuccessorExpression) {
			Element succElement = document.createElement("successor");
			Element subtermElement = document.createElement("subterm");
			succElement.appendChild(subtermElement);
			SuccessorExpression expr = (SuccessorExpression) expression;
			succElement.appendChild(parseArcExpression(expr.getSuccessorExpression(), document, subtermElement));
			structureElement.appendChild(succElement);
		}
		else if (expression instanceof PredecessorExpression) {
			Element predElement = document.createElement("predecessor");
			Element subtermElement = document.createElement("subterm");
			predElement.appendChild(subtermElement);
			PredecessorExpression expr = (PredecessorExpression) expression;
			predElement.appendChild(parseArcExpression(expr.getPredecessorExpression(), document, subtermElement));
			structureElement.appendChild(predElement);
		}
		else if (expression instanceof AddExpression) {
			Element addElement = document.createElement("add");
			AddExpression expr = (AddExpression) expression;
			for (ArcExpression arcExpression: expr.getAddExpression()) {
				Element subtermElement = document.createElement("subterm");
				addElement.appendChild(parseArcExpression(arcExpression,document,subtermElement));
			}
			structureElement.appendChild(addElement);
		}
		else if (expression instanceof SubtractExpression) {
			Element subtractElement = document.createElement("subtract");
			Element subtermElement = document.createElement("subterm");
			subtractElement.appendChild(subtermElement);
			SubtractExpression expr = (SubtractExpression) expression;
			subtractElement.appendChild(parseArcExpression(expr.getRightExpression(), document, subtermElement));
			subtractElement.appendChild(parseArcExpression(expr.getLeftExpression(), document, subtermElement));
			structureElement.appendChild(subtractElement);
		}
		else if (expression instanceof TupleExpression) {
			Element tupleElement = document.createElement("tuple");
			TupleExpression expr = (TupleExpression) expression;
			for (Expression colorExpressions : expr.getColors()) {
				Element subtermElement = document.createElement("subterm");
				tupleElement.appendChild(parseArcExpression(colorExpressions, document, subtermElement));
			}
			structureElement.appendChild(tupleElement);
		}
		return structureElement;
	}
	
	private Element createArcPoint(String x, String y, String type, Document document, int id) {
		Require.that(document != null, "Error: document was null");
		Element position = document.createElement("position");
		position.setAttribute("x", x);
		position.setAttribute("y", y);

		return position;
	}

	private void appendDeclarations(Document document, DataLayer guiModel, Element NET) {
		Require.that(guiModel != null, "Error: guiModel was null");
		Require.that(document != null, "Error: document was null");

		Element declarationElement = document.createElement("declaration");
		NET.appendChild(declarationElement);
		Element structureElement = document.createElement("structure");
		declarationElement.appendChild(structureElement);
		Element declarationsElement = document.createElement("declarations");
		structureElement.appendChild(parseSorts(document, declarationsElement));
	}

	private Element parseSorts(Document document, Element declarationsElement) {
		if(!network.colorTypes().isEmpty()) {
			for (ColorType colorType : network.colorTypes()) {
				if(colorType.getName().equals("dot")) {
					Element namedsortElement = document.createElement("namedsort");
					declarationsElement.appendChild(namedsortElement);
					namedsortElement.setAttribute("id", colorType.getId());
					namedsortElement.setAttribute("name", "Dot");
					Element dotElement = document.createElement("dot");
					namedsortElement.appendChild(dotElement);

				}
				else {
					Element namedsortElement = document.createElement("namedsort");
					declarationsElement.appendChild(namedsortElement);
					namedsortElement.setAttribute("id", colorType.getId());
					namedsortElement.setAttribute("name", colorType.getName());
					if (colorType instanceof ProductType) {
						Element productSortElement = document.createElement("productsort");
						namedsortElement.appendChild(productSortElement);
						for (ColorType ct : ((ProductType) colorType).getColorTypes()) {
							Element usersortElement = document.createElement("usersort");
							usersortElement.setAttribute("declaration", ct.getName());
							productSortElement.appendChild(usersortElement);
						}
					} else {
						Element cyclicElement = document.createElement("cyclicenumeration");
						namedsortElement.appendChild(cyclicElement);
						for (Color color : colorType) {
							Element feConstantElement = document.createElement("feconstant");
							feConstantElement.setAttribute("id", color.getColorName());
							feConstantElement.setAttribute("name", colorType.getName());
							cyclicElement.appendChild(feConstantElement);
						}
					}
				}
			}
		}
		for(Variable variable : network.variables()) {
			Element variableDeclearationElement = document.createElement("variabledecl");
			declarationsElement.appendChild(variableDeclearationElement);
			variableDeclearationElement.setAttribute("id", variable.getId());
			variableDeclearationElement.setAttribute("name", variable.getName());
			if(variable.getColorType().getName().equals("Dot")) {
				Element namedsortElement = document.createElement("namedsort");
				variableDeclearationElement.appendChild(namedsortElement);
				namedsortElement.setAttribute("id", variable.getId());
				namedsortElement.setAttribute("name", variable.getName());
				Element dotElement = document.createElement("dot");
				namedsortElement.appendChild(dotElement);
			}
			Element usersortElement = document.createElement("usersort");
			variableDeclearationElement.appendChild(usersortElement);
			usersortElement.setAttribute("declaration", variable.getColorType().getName());

		}

		return declarationsElement;
	}
}
